import os
import sys
import glob

from .loghandler import logger


def register_freecad(freecad_path: str = ""):
    """registers the freecad path and femtools in os.PATH

    Args:
        freecad_path (str): path to the local FreeCAD installation. If empty, will
            attempt to automagically find it

    Raises:
        ImportError: if the specified folder does not contain the FreeCAD Python libraries
    """
    supported_platforms = {
        "win32": [
            "C:/Program Files/FreeCAD *",
            os.path.expandvars("%LOCALAPPDATA%/Programs/FreeCAD *"),
        ],
        # "darwin": [] # macOS
        # "linux": [] # linux
    }
    if freecad_path is None or freecad_path == "":
        this_platform = sys.platform
        if this_platform not in supported_platforms.keys():
            try:
                raise ValueError(
                    f"Your platform ({this_platform}) is not yet explicitly supported. "
                    "You can still specify the path to FreeCAD manually using "
                    'parametric(freecad_path="path/to/freecad")'
                )
            except ValueError as e:
                logger.exception(str(e))
                raise
        for possible_path in supported_platforms[this_platform]:
            d = glob.glob(os.path.normpath(possible_path))
            if (len(d)) > 1:
                try:
                    raise ValueError(
                        "You seem to have multiple FreeCAD installations in your system. "
                        "You must specify the path to FreeCAD manually using "
                        'parametric(freecad_path="path/to/freecad")'
                    )
                except ValueError as e:
                    logger.exception(str(e))
                    raise

            if (len(d)) == 0:
                logger.debug(f"FreeCAD not found in {possible_path}")

            if (len(d)) == 1:
                freecad_path = d[0]
                logger.debug(f"Found FreeCAD at {freecad_path}")
                break

    if not freecad_path.endswith("bin"):
        freecad_path = os.path.join(freecad_path, "bin")

    if freecad_path not in sys.path:
        sys.path.append(os.path.normpath(freecad_path))
    # TODO: should automagically try to find freecad in the usual suspect folders;
    # it should also automatically add /bin if the user didn't specify it
    try:
        global FreeCAD, femtools, vtkResults
        FreeCAD = __import__("FreeCAD", globals(), locals())
        femtools = __import__("femtools.ccxtools", globals(), locals())
        vtkResults = __import__("feminout.importVTKResults", globals(), locals())
    except (ImportError, ModuleNotFoundError):
        logger.exception(f'"{freecad_path}" does not contain FreeCAD Python libraries')
        raise

    logger.debug(f"FreeCAD path added to sys.path: {freecad_path}")

    return (FreeCAD, femtools, vtkResults)



"""Provides a FreecadParametricFEA class to handle higher level parametric FEA functions,
    such as handling parameters and displaying results.
"""
import time
from typing import Union
from os import path
import pandas as pd
import numpy as np

import pickle

from tqdm import tqdm

import plotly.express as px

from .freecadmodel import FreecadModel
from .loghandler import logger


class parametric:
    """FreeCAD Parametric FEA object"""

    def __init__(self, freecad_path: str = "") -> None:
        """Initialises a FreecadParametricFEA object

        Args:
            freecad_path (str): (optional) path to the local FreeCAD installation.
                If not set, freecadmodel will make a few attempts at finding it in common
                locations

        """
        self.variables = []
        self.outputs = []
        self.fea_results_name = ""
        self.solver_name = ""

        self.freecad_path = freecad_path

        self.results_dataframe = pd.DataFrame()

        # initialise output headings to defaults
        self.set_outputs()

    def set_model(self, freecad_document: Union[str, FreecadModel]):
        """opens the freecad document and loads it

        Args:
            freecad_document (str or FreecadModel): path to the
                FreeCAD document, or a FreecadModel object
        """
        if isinstance(freecad_document, str):
            self.freecad_document = FreecadModel(
                document_path=freecad_document, freecad_path=self.freecad_path
            )
        elif isinstance(freecad_document, FreecadModel):
            self.freecad_document = freecad_document

        logger.info(f"FreeCAD document {freecad_document} loaded successfully")

    def set_variables(self, variables: list):
        """Sets the variables to run the batch analysis over.

        Args:
            variables (list of dict): a list of dictionaries. The dictionaries
                must contain:
                "object_name" (str): the object where the constraint is in,
                "constraint_name" (str): the name of the constraint to modify
                "constraint_values" (list of values):  values that the variable can assume
        """
        self.variables = variables

    def set_outputs(self, outputs: list = []):
        """Sets the variables to return as an output.

        Args:
            variables (list of dict): a list of dictionaries. The dictionaries
            must contain:
                "output_var" (str): the output variable (must be available in
                    the freecad fea results object),
                "reduction_fun" (function handle): a handle to the data
                    reduction function (e.g. np.max)
                ?"column_label" (str): (optional) label for the column.
                    Defaults to the function's __qualname__
        """
        if outputs == []:
            default_outputs = [
                {
                    "output_var": "vonMises",
                    "reduction_fun": np.max,
                },
                {
                    "output_var": "DisplacementLengths",
                    "reduction_fun": np.max,
                },
            ]
            self.outputs = default_outputs
        else:
            self.outputs = outputs

        logger.debug(f"Analysis outputs set to {self.outputs}")

    def setup_fea(self, fea_results_name: str, solver_name: str):
        """sets up the FEA analysis object

        Args:
            fea_results_name (str): name of the results object in the document
                e.g. CCX_Results
            solver_name (str): name of the solver object in the document
                e.g. SolverCcxTools
        """
        self.freecad_document.fea_results_name = fea_results_name
        self.freecad_document.solver_name = solver_name

    def run_parametric(
        self,
        dry_run: bool = False,
        export_results: bool = False,
        output_folder: str = "",
        quiet_mode: bool = False,
    ) -> pd.DataFrame:
        """runs the parametric sweep and returns the results

        Args:
            ?dry_run (bool): Doesn't run the FEA, but checks for model issues.
                Defaults to False
            ?export_results (bool): export results in .vtk format for each analysis
                Defaults to False
            ?output_folder (str): folder for results output
            ?quiet_mode (bool): suppresses all output.
                Defaults to False

        Returns:
            pd.DataFrame: Pandas dataframe containing the results
        """
        # TODO: this should let the user choose the type of run
        # e.g. "all" (full sampling), and other useful stuff like latin
        # hypercube sampling, random sampling, adaptive sampling...

        # change the target parameter in the CAD model.
        # as it stands it won't really support two parameters...
        # what if
        #  - create a dataframe with the parameters first
        #  - ran a single loop of all analyses over the dataframe
        #  - updated the dataframe?

        self.results_dataframe = self.populate_test_dataframe(
            self.variables, self.outputs
        )
        logger.debug("Results dataframe initialised")

        # iterate over all test cases

        if not quiet_mode:
            pbar = tqdm(total=len(self.results_dataframe), desc="Running test cases")

        for (test_case_idx, test_case_data) in self.results_dataframe.iterrows():
            # change each parameter to the value specified in the pd column:
            for parameter in self.variables:
                df_heading = self._param_to_df_heading(parameter)

                self.freecad_document.change_parameter(
                    object_name=parameter["object_name"],
                    constraint_name=parameter["constraint_name"],
                    target_value=test_case_data[df_heading],
                )

            # run (& time) the FEA
            if not dry_run:
                start_time = time.process_time()

                try:
                    fea_results_obj = self.freecad_document.run_fea()
                    fea_runtime = time.process_time() - start_time
                    logger.info(f"FEA test case {test_case_idx} ran in {fea_runtime}s")

                    if self.outputs == []:
                        self.set_outputs()
                    for output in self.outputs:
                        self.results_dataframe.loc[
                            test_case_idx, self._output_to_df_heading(output)
                        ] = output[  # type: ignore (looks like Pylance's fault)
                            "reduction_fun"
                        ](
                            fea_results_obj.getPropertyByName(output["output_var"])
                        )

                    self.results_dataframe.loc[
                        test_case_idx, "FEA_Runtime"
                    ] = fea_runtime  # type: ignore (looks like Pylance's fault)

                    # export if requested
                    # TODO: try and join the VTK files together as frames
                    if export_results:

                        (folder, filename) = path.split(self.freecad_document.filename)
                        (fn, _) = path.splitext(filename)

                        if output_folder != "":
                            folder = output_folder

                        n = int(
                            np.ceil(np.log10(len(self.results_dataframe) + 1))
                        )  # number of digits for vtk file

                        self.freecad_document.export_fea_results(
                            filename=path.join(
                                folder, f"FEA_{fn}_{test_case_idx:0{n}}.vtu"
                            ),
                            export_format="vtk",
                        )

                # TODO: may want to add runtime errors to the dataframe also
                # when in dry run
                except RuntimeError as e:
                    self.results_dataframe.loc[
                        test_case_idx, "Msg"
                    ] = str(  # type:ignore (looks like Pylance's fault)
                        e
                    )
                    logger.warning(f"Test case {test_case_idx} exited with error {e}")

            if not quiet_mode:
                pbar.update(1)  # type: ignore (only exists if quiet_mode is false)

        if not quiet_mode:
            pbar.close()  # type: ignore (only exists if quiet_mode is false)

        return self.results_dataframe

    def populate_test_dataframe(self, variables, outputs) -> pd.DataFrame:
        """Populates FreecadParametricFEA.results_dataframe with the
        test matrix to be run by the FEA batch. Uses self.variables
        and self.results.

        Args:
            variables (list): variables as defined in set_variables()

        Returns:
            pd.DataFrame: dataframe with test conditions and empty
                results columns
            param_headings: headings in the dataframe related to the variables
            output_headings: headings in the dataframe related to the output
        """
        param_vals = []
        param_headings = []
        output_headings = []

        for parameter in variables:
            param_vals.append(parameter["constraint_values"])
            param_headings.append(self._param_to_df_heading(parameter))

        for output in outputs:
            output_headings.append(self._output_to_df_heading(output))

        # Build list of n-param values
        grid = np.meshgrid(*param_vals)
        grid_list = list(x.ravel() for x in grid)

        df = pd.DataFrame()

        for (count, column) in enumerate(param_headings):
            df[column] = grid_list[count]

        for column in output_headings:
            df[column] = 0

        # generic empty data
        df["Msg"] = ""
        df["FEA_Runtime"] = 0
        logger.debug("Empty dataframe created")
        return df

    def plot_fea_results(self):
        """Plots the FEM analysis results using Plotly"""

        logger.debug("Preparing to plot FEA results")
        for output in self.outputs:
            if (len(self.variables)) == 1:
                fig = px.line(
                    self.results_dataframe,
                    x=self._param_to_df_heading(self.variables[0]),
                    y=self._output_to_df_heading(output),
                )
            elif (len(self.variables)) == 2:
                fig = px.line(
                    self.results_dataframe,
                    x=self._param_to_df_heading(self.variables[0]),
                    y=self._output_to_df_heading(output),
                    color=self._param_to_df_heading(self.variables[1]),
                )
            else:
                raise NotImplementedError(
                    "Plotting more than 2 variables not supported yet"
                )

            fig.show()

    def save_fea_results(self, results_filename: str, mode: str = "csv") -> None:
        """Saves the results of the analysis to a file.

        Args:
            results_filename (str): destination file
            mode (str, optional): saving mode. Can be one of:
                "csv" (default): comma separated values, as exported by Pandas
                "json": json file as exported by Pandas
                "pickle": .pickle file containing the Pandas dataframe

        Raises:
            NotImplementedError: if an export mode is not implemented.
        """
        if mode == "csv":
            self.results_dataframe.to_csv(results_filename)
        elif mode == "json":
            self.results_dataframe.to_json(
                results_filename, lines=True, orient="records"
            )
        elif mode == "pickle":
            with open(results_filename, "wb") as f:
                pickle.dump(self.results_dataframe, f)
        else:
            raise NotImplementedError(f"Export mode {mode} not yet implemented")

    def _param_to_df_heading(self, parameter) -> str:
        return f"{parameter['object_name']}.{parameter['constraint_name']}"

    def _output_to_df_heading(self, output) -> str:
        if "column_label" in output.keys():
            col_name = output["column_label"]
        else:
            col_name = output["reduction_fun"].__qualname__

        return f"{col_name}({output['output_var']})"



import logging
import logging.config

# taken from https://guicommits.com/how-to-log-in-python-like-a-pro/
ERROR_LOG_FILENAME = "./freecadparametricfea.log"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {  # The formatter name, it can be anything that I wish
            "format": "%(asctime)s:%(name)s:%(process)d:%(lineno)d " \
                      "%(levelname)s %(message)s",  # What to add in the message
            "datefmt": "%Y-%m-%d %H:%M:%S",  # How to display dates
        },
        "simple": {  # The formatter name
            "format": "%(message)s",  # As simple as possible!
        },
    },
    "handlers": {
        "logfile": {  # The handler name
            "formatter": "default",  # Refer to the formatter defined above
            "level": "INFO",
            "class": "logging.handlers.RotatingFileHandler",  # OUTPUT: Which class to use
            "filename": ERROR_LOG_FILENAME,
            "backupCount": 2,
        },
        "verbose_output": {  # The handler name
            "formatter": "simple",  # Refer to the formatter defined above
            "level": "DEBUG",  # FILTER: All logs
            "class": "logging.StreamHandler",  # OUTPUT: Which class to use
            "stream": "ext://sys.stdout",  # Stream to console
        },
    },
    "loggers": {
        "FreecadParametricFEA": {  # The name of the logger
            "level": "INFO",  # FILTER: only INFO logs onwards
            "handlers": [
                "logfile",  # Refer the handler defined above
            ],
        },
    },
}

logging.config.dictConfig(LOGGING_CONFIG)
logger = logging.getLogger(__name__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['freecadparametricfea']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.24.1,<2.0.0',
 'pandas>=1.5.2,<2.0.0',
 'plotly>=5.12.0,<6.0.0',
 'pyside2>=5.15.2.1,<6.0.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'freecadparametricfea',
    'version': '0.3',
    'description': 'A flexible parametric FEA library based on FreeCAD',
    'long_description': '# freecadparametricfea\n\n A flexible parametric FEA library based on [FreeCAD](https://www.freecadweb.org/), currently supporting FreeCAD 0.20 on Windows.\n \n If you have 20 minutes I recommend the video tutorial on the [@engineeringmaths Youtube channel](https://www.youtube.com/watch?v=cwtgB4KpdJo).\n\n> **Warning:**\n> this project is very early release, and should not be used for any serious structural analysis. It is aimed at hobbyists and makers\n\n## Quickest start\nCreate a Python 3.8 virtual environment:\n\n`pipenv --python 3.8`\n\nInstall the latest version from pypi:\n\n`pipenv install freecadparametricfea`\n\nthen run any of the examples inside [the examples folder](examples/)\n\n## Quick start\n\nCreate a FreeCAD part and assign names to the constraints that you want to change. You need to set up a FEA analysis as well, I have tested this using CalculiX and Netgen.\n\nThen in a script, or on the command line, run:\n\n```python\nfrom FreecadParametricFEA import parametric\nimport numpy as np\n\n# initialise a parametric FEA object\nfea = parametric()\n\n# load the FreeCAD model\nfea.set_model("your-part-here.fcstd")\n\n# list the parameters to sweep:\nfea.set_variables(\n    [\n        {\n            "object_name": "CutsSketch", # the object where to find the constraint\n            "constraint_name": "NotchDistance", # the constraint name that you assigned \n            "constraint_values": np.linspace(10, 30, 5), # the values you want to check\n        },\n        {\n            "object_name": "CutsSketch",\n            "constraint_name": "NotchDiam",\n            "constraint_values": np.linspace(5, 9, 5),\n        },\n    ]\n)\n\n# run and save the results (will return a Pandas DataFrame)\nresults = fea.run_parametric()\n\n# plot the results\nfea.plot_fea_results()\n```\n\n## Feeling fancy\n\n### Custom outputs\nThe default is to export the max Von Mises stress and max displacement values. You can also specify your own values and data reduction function like this:\n\n```python\nfea.set_outputs([\n        {\n            "output_var": "vonMises",\n            "reduction_fun": np.median,\n        },\n        {\n            "output_var": "vonMises",\n            "reduction_fun": lambda v: np.percentile(v, 95),\n            "column_label": "95th percentile"\n        }\n    ])\n```\n\n### Changing materials\nYou can specify any material that you can find in the FreeCAD FEA material selection dropdown; just refer to it by its name:\n\n```python\nfea.set_outputs([\n    {\n        "object_name": "MaterialSolid",  # the object where to find the constraint\n        "constraint_name": "Material",  # the constraint name that you assigned\n        "constraint_values": ["Aluminium-Generic", "Steel-Generic"],\n    },\n])\n```\n### Different names for CCX solver and CCX results\nRenaming the CCX solver and results won\'t affect the solution, but if you\'re having trouble running the analysis you can set them yourself just before `run_parametric()`:\n\n```python\n# in case you need to explicitly set the CalculiX results object and the solver name\nfea.setup_fea(fea_results_name="CCX_Results", solver_name="SolverCcxTools")\n```\n\n### Exporting data\n\nYou can export individual ParaView files using:\n\n```python\nresults = fea.run_parametric(export_results=True, output_folder="path/to/my/results")\n```\n\n\nOr just save the results dataframe in a .csv, json or serialised pickle object:\n\n```python\nfea.save_fea_results("results.csv")\nfea.save_fea_results("results.json", mode="json")\nfea.save_fea_results("results.pickle", mode="pickle")\n```\n\n... or even take a look at the parameters matrix before running any analysis:\n\n```python\nresults = fea.run_parametric(dry_run=True)\n```\n\n### Custom FreeCAD path\nIf you have multiple installations of FreeCAD or are using a system other than Windows (as of version <=0.3) you have to specify the path to FreeCAD manually in the call to `parametric`:\n\n```python\n# you can manually specify the path to FreeCAD on your system:\nFREECAD_PATH = "C:/Program Files/FreeCAD 0.20/bin"\nfea = parametric(freecad_path=FREECAD_PATH)\n```\n# Limitations and caveats\n\nAs of 0.3:\n * this has been tested on FreeCAD 0.20, on Windows only, but you can try other platforms\n * only Netgen meshes are supported\n * Only static FEM analysis has been tested\n\n# Contributing\nI have created this for hobby and personal use, as I was interested in learning more about FreeCAD and writing Python modules. There are a lot of things that I would like to fix, if you want to get involved have a look at the [open issues](https://github.com/da-crivelli/freecad-parametric-fea/issues/) and send me a message if you have any questions.\n\n\n',
    'author': 'Davide Crivelli,',
    'author_email': 'da.crivelli@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/da-crivelli/freecad-parametric-fea',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.10,<3.9.0',
}


setup(**setup_kwargs)



"""FreecadModel object and helpers"""
import os
import contextlib
from .loghandler import logger

from .register_freecad import register_freecad

from typing import Tuple


class FreecadModel:
    """FreecadModel class"""

    def __init__(self, document_path: str, freecad_path: str = "") -> None:
        """initialises a FreecadModel object

        Args:
            document_path (str): path to the FreeCAD file
            freecad_path (str): path to the FreeCAD Python libraries
        """
        self.filename = document_path

        global FreeCAD, femtools, vtkResults
        (FreeCAD, femtools, vtkResults) = register_freecad(freecad_path=freecad_path)

        self.model = FreeCAD.open(document_path)
        logger.debug(f"Opened FreeCAD model {document_path}")

        self.solver_name = ""
        self.fea_results_name = ""
        # TODO: error handling

    def change_parameter(
        self, object_name: str, constraint_name: str, target_value: float
    ):
        """changes a parameter (e.g. a named constraint) inside a freecad
        document. Currently works if the constraint is inside the driving
        sketch of the referenced object (e.g. a pocket)

        Args:
            object_name (str): name of the Freecad object containing the
                sketch containing the constraint
            constraint_name (str): name of the constraint to modify
            target_value (float): target value for the constraint
        """

        target_object = self.model.getObjectsByLabel(object_name)
        if not target_object:
            try:
                raise KeyError(f"Unable to find object {object_name} in the model")
            except KeyError as e:
                logger.exception(str(e))
                raise

        # supported specific object types:
        # <Sketcher::SketchObject>
        # <App::MaterialObjectPython object>

        # FIXME: this should really be done via type checking

        try:
            target_str = str(target_object[0])

            # sketcher objects need obj.getDatum / obj.setDatum
            if target_str == "<Sketcher::SketchObject>":
                target_object[0].getDatum(constraint_name)
                logger.debug(
                    f"Object {object_name} is a sketch, found {constraint_name}"
                )

                target_object[0].setDatum(constraint_name, target_value)

            # materials need some special treatment via material cards
            elif target_str == "<App::MaterialObjectPython object>":
                from materialtools.cardutils import import_materials as getmats  # type: ignore

                materials, _, _ = getmats(target_object[0].Category)

                for (_, m) in materials.items():
                    if m["CardName"] == target_value:
                        target_object[0].Material = m
                        return

                logger.debug(
                    f"Object {object_name} is a material, "
                    "setting material {constraint_value}"
                )

            # generic objects need setattr(obj, attr, value)
            else:
                getattr(target_object[0], constraint_name)
                logger.debug(
                    f"Object {object_name} is an object, found {constraint_name}"
                )

                setattr(target_object[0], constraint_name, target_value)

        except (NameError, IndexError):
            logger.exception(
                f"Invalid constraint name {constraint_name} in object {object_name}"
            )
            raise

        logger.debug(f"Set {object_name}.{constraint_name} to {target_value}")
        # apply changes and recompute
        self.model.recompute()
        logger.debug("Model recomputed")
        # TODO: check for model errors here

    def run_fea(self):
        """runs a FEA analysis in the specified freecad document

        Args:
            solver_name (str): name of the solver (e.g. SolverCcxTools)
            fea_results_name (str): name of the FEA results container (e.g.
                CCX_Results)

        Returns:
            fea object: a FreeCAD object containing the FEA results
        """

        if self.solver_name == "":
            self._find_solver_result_names()

        solver_object = self.model.getObject(self.solver_name)

        fea = femtools.ccxtools.FemToolsCcx(solver=solver_object)
        fea.purge_results()
        fea.reset_all()
        fea.update_objects()
        logger.debug(f"Prepared solver {solver_object.Name}")

        # there should be some error handling here
        fea.check_prerequisites()
        logger.debug("Checked FEA prerequisites")
        # patching this because Calculix prints some useless info in
        # Freecad 0.20 for solid models only... see bug #3
        with open(os.devnull, "w", encoding="utf8") as devnull:
            with contextlib.redirect_stdout(devnull):
                fea.run()

        if fea.results_present:
            logger.debug("FEA results generated")
            return self.model.getObject(self.fea_results_name)
        else:
            try:
                raise RuntimeError("FEA results are not present")
            except RuntimeError as e:
                logger.exception(str(e))
                raise

    def export_fea_results(self, filename: str, export_format: str = "vtk"):
        """exports the results of a analysis to various mesh formats

        Args:
            filename (str): path to the output file
            export_format (str, optional): output format. Defaults to "vtk".

        Raises:
            NotImplementedError: if the output format specified is not available
        """

        if export_format == "vtk":
            objects = []
            objects.append(self.model.getObject(self.fea_results_name))
            vtkResults.importVTKResults.export(objects, filename)
            logger.info(f"Exporting VTK file {filename}")
            del objects
        else:
            try:
                raise NotImplementedError(
                    f"Export method {export_format} not available"
                )
            except NotImplementedError as e:
                logger.exception(str(e))
                raise

    def _find_solver_result_names(self) -> Tuple[str, str]:
        # do stuff...
        solver_name = ""
        fea_results_name = ""

        for el in self.model.Objects:
            if str(el) == "<Fem::FemSolverObjectPython object>":
                solver_name = el.Name
            if str(el) == "<Fem::FemResultObjectPython object>":
                fea_results_name = el.Name

        if "" in (solver_name, fea_results_name):
            try:
                raise NameError(
                    "FEA solver or results not found, consider "
                    "specifying manually using parametric.setup_fea()"
                )
            except NameError as e:
                logger.exception(str(e))
                raise

        logger.debug(f"Found solver name: {solver_name}")
        logger.debug(f"Found FEA results name: {fea_results_name}")

        self.solver_name = solver_name
        self.fea_results_name = fea_results_name
        return (self.solver_name, self.fea_results_name)
